import {readonly, ref} from 'vue'
import {defineStore, acceptHMRUpdate} from 'pinia'

import ListModel from '@/models/list'
import {checkAndSetApiUrl} from '@/helpers/checkAndSetApiUrl'

import {useAuthStore} from '@/stores/auth'
import type {IList} from '@/modelTypes/IList'

export const useBaseStore = defineStore('base', () => {
	const loading = ref(false)
	const ready = ref(false)

	// This is used to highlight the current list in menu for all list related views
	const currentList = ref<IList | null>(new ListModel({
		id: 0,
		isArchived: false,
	}))

	const hasTasks = ref(false)
	const menuActive = ref(true)
	const keyboardShortcutsActive = ref(false)
	const quickActionsActive = ref(false)
	const logoVisible = ref(true)

	function setLoading(newLoading: boolean) {
		loading.value = newLoading
	}

	function setCurrentList(newCurrentList: IList | null) {
		// Server updates don't return the right. Therefore, the right is reset after updating the list which is
		// confusing because all the buttons will disappear in that case. To prevent this, we're keeping the right
		// when updating the list in global state.
		if (
			typeof currentList.value?.maxRight !== 'undefined' &&
			newCurrentList !== null &&
			(
				typeof newCurrentList.maxRight === 'undefined' ||
				newCurrentList.maxRight === null
			)
		) {
			newCurrentList.maxRight = currentList.value.maxRight
		}
		currentList.value = newCurrentList
	}

	function setHasTasks(newHasTasks: boolean) {
		hasTasks.value = newHasTasks
	}

	function setMenuActive(newMenuActive: boolean) {
		menuActive.value = newMenuActive
	}

	function toggleMenu() {
		menuActive.value = !menuActive.value
	}

	function setKeyboardShortcutsActive(value: boolean) {
		keyboardShortcutsActive.value = value
	}

	function setQuickActionsActive(value: boolean) {
		quickActionsActive.value = value
	}

	function setLogoVisible(visible: boolean) {
		logoVisible.value = visible
	}
	
	function setReady(value: boolean) {
		ready.value = value
	}

	function handleSetCurrentList(
		{list}: {list: IList | null},
	) {
		if (list === null) {
			setCurrentList({})
			return
		}

		setCurrentList(list)
	}

	const authStore = useAuthStore()
	async function loadApp() {
		await checkAndSetApiUrl(window.API_URL)
		await authStore.checkAuth()
		ready.value = true
	}

	return {
		loading: readonly(loading),
		ready: readonly(ready),
		currentList: readonly(currentList),
		hasTasks: readonly(hasTasks),
		menuActive: readonly(menuActive),
		keyboardShortcutsActive: readonly(keyboardShortcutsActive),
		quickActionsActive: readonly(quickActionsActive),
		logoVisible: readonly(logoVisible),

		setLoading,
		setCurrentList,
		setHasTasks,
		setMenuActive,
		toggleMenu,
		setKeyboardShortcutsActive,
		setQuickActionsActive,
		setLogoVisible,
		setReady,

		handleSetCurrentList,
		loadApp,
	}
})

// support hot reloading
if (import.meta.hot) {
  import.meta.hot.accept(acceptHMRUpdate(useBaseStore, import.meta.hot))
}